import java.util.*;
import java.io.*;

/**
 * This class represents a finite automaton in both
 * non-determinstic and determinstic variants.
 *
 */
public class Automaton {
    // Tabular representation of automaton transition matrix
    private Set[][] automaton;
    private int numStates, numLabels;
    // Automaton consists of states and labels, transitions are managed by states
    private Set states, labels;
    private State currentState;
    // Utility data structure to map unique names to corresponding instance
    private Map nameToState, nameToLabel;
    // Utility data structure to map integer id to states and labels
    private List stateList, labelList;

    // Teilaufgabe a) ---------------------------------------------------------

    // Returns a boolean matrix with marked equivalent state pairs
    private boolean[][] calculateEquivalenceMatrix() {
        boolean[][] marked = new boolean[numStates][numStates];
        // Use lower triangle of matrix as equivalence is reflexive and symmetric
        for (int i = 0; i < numStates; i++)
            for (int j = 0; j < i; j++)
                marked[i][j] = false;
        // Init: set homogeneous terminal/non-terminal state pairs as equivalent
        for (int i = 0; i < numStates; i++) {
            for (int j = 0; j < i; j++) {
                State si = (State) stateList.get(i);
                State sj = (State) stateList.get(j);
                if ((si.isTerminal() && sj.isTerminal()) ||
                    (!si.isTerminal() && !sj.isTerminal()))
                    marked[i][j] = true;
            }
        }
        // Elongation: check labels and state pairs for possible transitions
        boolean isStable = false;
        while (!isStable) {
            isStable = true;
            for (int i = 0; i < numStates; i++) {
                for (int j = 0; j < i; j++) {
                    if (marked[i][j]) {
                        Iterator labelIt = labels.iterator();
                        while (labelIt.hasNext()) {
                            Label l = (Label) labelIt.next();
                            Set rm = automaton[i][l.getId()];
                            Set rn = automaton[j][l.getId()];
                            // Automaton deterministic & total -> well-defined result
                            State sm = (State) rm.iterator().next();
                            State sn = (State) rn.iterator().next();
                            int m = Math.max(sm.getId(), sn.getId());
                            int n = Math.min(sm.getId(), sn.getId());
                            // Check for transition into non-equivalent state pair,
                            // excluding reflexive pairs (which are always equivalent)
                            if (!marked[m][n] && m != n) {
                                marked[i][j] = false;
                                isStable = false;
                            }
                        }
                    }
                }
            }
        }
        return marked;
    }

    // Teilaufgabe b) ---------------------------------------------------------

    // Minimize automaton from given matrix of equivalent state pairs
    private void fromEquivalenceMatrix(boolean[][] matrix) {
        System.out.println(matrixToString(matrix));
        Set removed = new HashSet();
        for (int i = 0; i < numStates; i++) {
            for (int j = 0; j < i; j++) {
                if (matrix[i][j]) {
                    // Marked state pairs are equivalent
                    State si = (State) stateList.get(i);
                    State sj = (State) stateList.get(j);
                    // Both states still relevant?
                    if (!removed.contains(si) && !removed.contains(sj)) {
                        System.out.println("Merging " + si + " and " + sj +
                                           " [" + i + "," + j + "]");
                        // Redirect incoming transitions of si to sj
                        Iterator incomingIt = si.getIncomingTransitions().iterator();
                        while (incomingIt.hasNext()) {
                            String from;
                            Transition t = (Transition) incomingIt.next();
                            // Check for loop transitions
                            if (t.getOrigin() != t.getTarget())
                                from = t.getOrigin().getName();
                            else
                                from = sj.getName();
                            System.out.println("Redirecting " + t + " from " + from +
                                               " to " + sj + " (instead of " + si + ")");
                            addTransition(from, sj.getName(), t.getLabel().getName());
                        }
                        // Transfer start property if necessary
                        if (si.isStart())
                            sj.setStart(true);
                        // Schedule processed state for later removal
                        System.out.println("Removing " + si);
                        removed.add(si);
                    }
                }
            }
        }
        // Delete processed states
        Iterator remIt = removed.iterator();
        while (remIt.hasNext()) {
            State s = (State) remIt.next();
            removeState(s.getName());
        }
    }

    // Teilaufgabe c) ---------------------------------------------------------
    public void minimize() {
        fromEquivalenceMatrix(calculateEquivalenceMatrix());
        init();
    }

    // Vorgegebener Teil -------------------------------------------------------

    public Automaton() {
        numStates = 0; numLabels = 0;
        states = new HashSet(); labels = new HashSet();
        nameToState = new HashMap(); nameToLabel = new HashMap();
        stateList = new Vector(); labelList = new Vector();
    }

    // Add a new state to the automaton, possibly as start or terminal state
    public void addState(String name, boolean isStart, boolean isTerminal) {
        State s = (State) nameToState.get(name);
        if (s == null) {
            s = new State(name, numStates);
            s.setStart(isStart); s.setTerminal(isTerminal);
            states.add(s);
            stateList.add(s);
            nameToState.put(name, s);
            numStates++;
        }
    }

    // Remove a state from the automaton if it already exists
    public void removeState(String name) {
        State s = (State) nameToState.get(name);
        if (s != null) {
            // Remove outgoing and incoming transition
            Iterator transitions = s.getOutgoingTransitions().iterator();
            while (transitions.hasNext()) {
                Transition t = (Transition) transitions.next();
                t.getTarget().getIncomingTransitions().remove(t);
            }
            transitions = s.getIncomingTransitions().iterator();
            while (transitions.hasNext()) {
                Transition t = (Transition) transitions.next();
                t.getOrigin().getOutgoingTransitions().remove(t);
            }
            // Remove state itself
            states.remove(s);
            nameToState.remove(name);
            stateList.remove(s);
            numStates--;
        }
    }

    // Add a new transition to the automaton between possibly existing states
    public void addTransition(String from, String to, String label) {
        State origin = (State) nameToState.get(from);
        State target = (State) nameToState.get(to);
        if (origin == null) {
            addState(from, false, false);
            origin = (State) nameToState.get(from);
        }
        if (target == null) {
            addState(to, false, false);
            target = (State) nameToState.get(to);
        }
        // Create label if necessary
        Label l = (Label) nameToLabel.get(label);
        if (l == null) {
            l = new Label(label, numLabels);
            labels.add(l);
            labelList.add(l);
            nameToLabel.put(label, l);
            numLabels++;
        }
        // Create transition itself if not already present
        Transition t = new Transition(origin, target, l);
        Iterator tit = origin.getOutgoingTransitions().iterator();
        while (tit.hasNext()) {
            Transition old = (Transition) tit.next();
            if (t.equals(old)) return;
        }
        origin.getOutgoingTransitions().add(t);
        target.getIncomingTransitions().add(t);
    }

    // Remove a transition from the automaton between existing states
    public void removeTransition(String from, String to, String label) {
        State origin = (State) nameToState.get(from);
        State target = (State) nameToState.get(to);
        Label l = (Label) nameToLabel.get(label);
        if (origin == null || target == null || l == null) {
            System.out.println("-> Origin/target state or label not found!");
        } else {
            // Remove transition and update internal structure
            Iterator transitions = origin.getOutgoingTransitions().iterator();
            while (transitions.hasNext()) {
                Transition t = (Transition) transitions.next();
                if (t.getTarget() == target && t.getLabel() == l) {
                    origin.getOutgoingTransitions().remove(t);
                    target.getIncomingTransitions().remove(t);
                    return;
                }
            }
        }
    }

    // Initialize the automaton, called before execution
    public void init() {
        initTransitionMatrix();
        // Choose an arbitrary initial state
        Iterator sit = states.iterator();
        while (sit.hasNext()) {
            State s = (State) sit.next();
            if (s.isStart()) {
                currentState = s;
                break;
            }
        }
        if (currentState == null)
            System.out.println("-> No initial state present!");
    }

    // Construct automaton from given regular grammar with productions of form
    // <nonterminal> -> <epsilon>|terminal|terminal <nonterminal>
    public void readGrammar(String fileName) {
        // Initialize private fields
        numStates = 0; numLabels = 0; currentState = null;
        nameToState = new HashMap(); nameToLabel = new HashMap();
        states = new HashSet(); labels = new HashSet();
        stateList = new Vector(); labelList = new Vector();

        // Iterate all productions in given grammar (no error checking)
        String buffer = new String();
        String lhs = new String(), rhs1 = new String(), rhs2 = new String();
        try {
            BufferedReader reader = new BufferedReader(new FileReader(fileName));
            while (reader.ready()) {
                // Separate productions into left- and right-hand side parts
                buffer = reader.readLine();
                StringTokenizer tokenizer = new StringTokenizer(buffer);
                lhs = tokenizer.nextToken();
                tokenizer.nextToken();
                rhs1 = tokenizer.nextToken();
                if (tokenizer.hasMoreTokens())
                    rhs2 = tokenizer.nextToken();
                else
                    rhs2 = "";

                // Process originating state i.e. nonterminal on left-hand side
                addState(lhs, isStart(lhs), false);
                if (isEpsilon(rhs1)) {
                    // Production: '<nonterminal> -> 'epsilon>'
                    State origin = (State) nameToState.get(lhs);
                    origin.setTerminal(true);
                } else {
                    // Process target state if necessary
                    if (rhs2.length() > 0) {
                        // Production: '<nonterminal> -> terminal <nonterminal>'
                        addState(rhs2, false, false);
                        addTransition(lhs, rhs2, rhs1);
                    } else {
                        // Production: '<nonterminal> -> terminal'
                        // Add dedicated final state
                        addState("<F>", false, true);
                        addTransition(lhs, "<F>", rhs1);
                    }
                }
            }
            reader.close();
        }
        catch (IOException ex) {
            ex.printStackTrace();
            System.exit(1);
        }
    }

    // Initialize transition matrix after change to automaton structure
    private void initTransitionMatrix() {
        automaton = new Set [numStates][numLabels];
        for (int i = 0; i < numStates; i++) {
            for (int j = 0; j < numLabels; j++) {
                automaton[i][j] = new HashSet();
            }
        }
        // Update integer id after possible previous removal of states
        for (int i = 0; i < numStates; i++) {
            State s = (State) stateList.get(i);
            s.setId(i);
        }
        // For all states: add target state of each outgoing transition to
        // result set for its label
        Iterator sit = states.iterator();
        while (sit.hasNext()) {
            State s = (State) sit.next();
            Iterator tit = s.getOutgoingTransitions().iterator();
            while (tit.hasNext()) {
                Transition t = (Transition) tit.next();
                automaton[s.getId()][t.getLabel().getId()].add(t.getTarget());
            }
        }
    }

    // Detect a number of special tokens
    private boolean isEpsilon(String token) {
        if (token.equals("<epsilon>"))
            return true;
        return false;
    }
    private boolean isStart(String token) {
        if (token.equals("<S>"))
            return true;
        return false;
    }

    // Formatted output of automaton
    public String toString() {
        String result = new String();
        result += "States: ";
        Iterator sit = states.iterator();
        while (sit.hasNext()) {
            State s = (State) sit.next();
            result += s + ":" + s.getId();
            if (s.isStart()) result += "[s]";
            if (s.isTerminal()) result += "[f]";
            if (sit.hasNext()) result += ", ";
        }
        result += "\nLabels: ";
        Iterator lit = labels.iterator();
        while (lit.hasNext()) {
            result += lit.next();
            if (lit.hasNext()) result += ", ";
        }
        result += "\nTransitions:\n";
        sit = states.iterator();
        while (sit.hasNext()) {
            State s = (State) sit.next();
            Iterator tit = s.getOutgoingTransitions().iterator();
            while (tit.hasNext()) {
                Transition t = (Transition) tit.next();
                result += "\t" + s + " -" + t + "-> " + t.getTarget() +"\n";
            }
        }
        result += "\n";
        return result;
    }

    // Formated output of a boolean matrix
    private String matrixToString(boolean[][] m) {
        String result = new String("\t|");
        for (int j = 0; j < m.length; j++) {
            result += j + "\t|";
        }
        result += "\n";
        for (int i = 0; i < m.length; i++) {
            result += i + "\t|";
            for (int j = 0; j < m.length; j++) {
                if (m[i][j]) result += "x\t|"; else result += "\t|";
            }
            result += "\n";
        }
        return result;
    }

}
